from __future__ import annotations

import os
from typing import Optional, Dict, Any, List, Union
from dataclasses import dataclass

from ._api_client import Client, AuthenticatedClient
from .active_run import ActiveRun
from .active_artifact import ActiveArtifact
from .model import ArtifactRef, ModelStage


@dataclass
class ConnectionOptions:
    """Specify options for a MLAideClient"""

    server_url: Optional[str]
    api_key: Optional[str]

    def __init__(self, server_url: str = None, api_key: str = None):
        self.server_url = server_url
        self.api_key = api_key

    def to_dict(self) -> Dict[str, Any]:
        d = {
            "server_url": self.server_url,
            "api_key": self.api_key
        }

        # Remove values from dict that are None
        return {k: v for k, v in d.items() if v is not None}

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> ConnectionOptions:
        if d is None:
            d = dict()

        options = ConnectionOptions(
            server_url=d.get("server_url", None),
            api_key=d.get("api_key", None)
        )

        return options


class MLAideClient:
    """This is the main entry point to use this library. Creates a connection to the ML Aide server and provides
    read and write access to all resources.
    """

    __options: ConnectionOptions
    __api_client: Client
    __project_key: str

    def __init__(self, project_key: str, options: ConnectionOptions = None):
        """Creates a new instance of this class.

        Arguments:
            project_key: The key of the project, that should be accessed. All operations will be made on this project.
            options: Optional options that will be used to establish a connection.
        """
        if project_key is None:
            raise ValueError("project key must be not None")
        self.__project_key = project_key

        if options is None:
            self.__options = MLAideClient.__get_default_options()
        else:
            self.__options = MLAideClient.__merge_options(MLAideClient.__get_default_options(), options)

        self.__api_client = AuthenticatedClient(base_url=self.__options.server_url,
                                                api_key=self.__options.api_key)

    def start_new_run(self,
                      experiment_key: str = None,
                      run_name: str = None,
                      used_artifacts: List[ArtifactRef] = None,
                      auto_create_experiment: bool = True) -> ActiveRun:
        """Creates and starts a new run, that will be assigned to the specified experiment. The run object can be used
        to log all necessary information.

        Arguments:
            experiment_key: The key of the experiment, that the new run should be assigned to. If `None` a new, random
            experiment will be created.
            run_name: The name of the run. The name helps to identify the run for humans. If `None` a random name will
            be used.
            used_artifacts: An optional list of `ArtifactRef` that references artifacts, that are used as input for
            this run. This information will help to create and visualize the experiment lineage.
            auto_create_experiment: Specifies whether the experiment (see `experiment_key`) should be created if it
            does not exist or not. If `auto_create_experiment` is `False` and the experiment does not exist an error
            will be raised.

        Returns:
            This object encapsulates the newly created run and provides functions to log all information \
            that belongs to the run.
        """
        return ActiveRun(self.__api_client,
                         self.__project_key,
                         experiment_key,
                         run_name,
                         used_artifacts,
                         auto_create_experiment)

    def get_artifact(self, name: str, version: Optional[int] = None) -> ActiveArtifact:
        """Gets an existing artifact. The artifact is specified by its name and version. If no version
        is specified, the latest available version of the artifact will be used.

        Arguments:
            name: The name of the artifact.
            version: The (optional) version of the artifact. If no version is specified, the latest available version
            will be loaded.

        Returns:
             This object encapsulates an artifact and provides functions to interact with the artifact.
        """

        return ActiveArtifact(self.__api_client, self.__project_key, name, version)

    def load_model(self,
                   name: str,
                   version: Optional[int] = None,
                   stage: Optional[ModelStage] = None) -> any:
        """Loads and restores a model. The model is specified by its name and version. If no version
        is specified, the latest available version of the model will be used.

        Arguments:
            name: The name of the model.
            version: The (optional) version of the model. If no version is specified, the latest available version
            will be loaded.
            stage: This argument can only be used when version is None. In this case the latest model can be filtered
            by its stage. In reverse this means that all model versions will be ignored when they have not the specified
            stage.

        Returns:
             The model. E.g. in the case of a scikit-learn model the return value will be a deserialized model that
             can be used for predictions using `.predict(...)`.
         """

        if version is not None and stage is not None:
            raise ValueError("Only one argument of version and stage can be not None")

        return ActiveArtifact(self.__api_client, self.__project_key, name, version, stage).load_model()

    @property
    def options(self) -> ConnectionOptions:
        return self.__options

    @property
    def api_client(self) -> Client:
        return self.__api_client

    @staticmethod
    def __get_default_options() -> ConnectionOptions:
        options = ConnectionOptions()
        options.server_url = 'http://localhost:9000/api/v1'
        options.api_key = os.environ.get('MLAIDE_API_KEY')
        return options

    @staticmethod
    def __merge_options(target: ConnectionOptions, source: ConnectionOptions) -> ConnectionOptions:
        merged = dict()
        merged.update(target.to_dict())
        merged.update(source.to_dict())

        return ConnectionOptions.from_dict(merged)



from . import mapper, _model_deser
from ._api_client import Client
from ._api_client.api import artifact_api
from .model import Artifact, ModelStage

from dataclasses import replace
from io import BytesIO
from typing import Optional, Tuple
from zipfile import ZipFile


class ActiveArtifact(object):
    """This class provides access to artifacts that are stored in ML Aide"""

    __api_client: Client
    __project_key: str
    __artifact: Artifact
    __cached_zip: Tuple[BytesIO, Optional[str]] = None

    def __init__(self,
                 api_client: Client,
                 project_key: str,
                 artifact_name: str,
                 artifact_version: Optional[int],
                 model_stage: Optional[ModelStage] = None):
        self.__api_client = api_client
        self.__project_key = project_key
        self.__artifact = self.__load_artifact(artifact_name, artifact_version, model_stage)

    def __load_artifact(self, artifact_name: str, artifact_version: Optional[int],
                        model_stage: Optional[ModelStage]) -> Artifact:
        artifact_dto = artifact_api.get_artifact(client=self.__api_client,
                                                 project_key=self.__project_key,
                                                 artifact_name=artifact_name,
                                                 artifact_version=artifact_version,
                                                 model_stage=model_stage.value if model_stage is not None else None)
        return mapper.dto_to_artifact(artifact_dto)

    @property
    def artifact(self) -> Artifact:
        # Return a deep copy to avoid changing anything by the client
        return replace(self.__artifact)

    def load(self, filename: str) -> BytesIO:
        """Load a specific file of this artifact into memory

        Arguments:
            filename: The name of the file that should be loaded
        """

        # TODO: Do not download whole zip; instead download just the single file
        zip_bytes, zip_filename = self.__download_zip()
        with ZipFile(zip_bytes) as z:
            zip_info = z.infolist()
            desired_file = next(info for info in zip_info if info.filename == filename)
            with z.open(desired_file, 'r') as zip_file:
                return BytesIO(zip_file.read())

    def load_model(self):
        model_binary = self.load("model.pkl")
        return _model_deser.deserialize(model_binary)

    def download(self, target_directory: str):
        """Downloads all files of this artifact and stores them into the specified directory.

        Arguments:
            target_directory: The path to the directory where all files should be stored.
        """

        # download
        artifact_bytes, artifact_filename = self.__download_zip()

        # unzip and write to disk
        with ZipFile(artifact_bytes) as z:
            z.extractall(target_directory)

    def __download_zip(self) -> (BytesIO, str):
        if self.__cached_zip is None:
            self.__cached_zip = artifact_api.download_artifact(client=self.__api_client,
                                                               project_key=self.__project_key,
                                                               artifact_name=self.__artifact.name,
                                                               artifact_version=self.__artifact.version)

        return self.__cached_zip



from datetime import datetime
from dataclasses import dataclass
from enum import Enum
from typing import Collection, Dict, Optional
from . import UserRef


class ModelStage(Enum):
    NONE = 'NONE',
    STAGING = 'STAGING',
    PRODUCTION = 'PRODUCTION',
    ABANDONED = 'ABANDONED',
    DEPRECATED = 'DEPRECATED'


@dataclass
class ModelRevision(object):
    new_stage: ModelStage
    old_stage: ModelStage
    created_at: Optional[datetime] = None
    created_by: Optional[UserRef] = None
    note: Optional[str] = None


@dataclass
class Model(object):
    created_at: Optional[datetime] = None
    created_by: Optional[UserRef] = None
    model_revisions: Optional[Collection[ModelRevision]] = None
    metadata: Optional[Dict[str, str]] = None
    stage: Optional[ModelStage] = None
    updated_at: Optional[datetime] = None



from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ExperimentRefDto:
    experiment_key: str



from .user_ref import UserRef
from .artifact_file import ArtifactFile
from .artifact_ref import ArtifactRef
from .model import Model, ModelRevision, ModelStage
from .run import Run, RunStatus
from .artifact import Artifact



from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined
from typing import Union


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ArtifactRefDto(object):
    name: str
    version: Union[int, str, None] = None



from dataclasses import field
from dataclasses_json import config, DataClassJsonMixin
from datetime import datetime
from dateutil.parser import isoparse
from marshmallow import fields
from typing import Optional


def optional_iso_datetime_encoder(value: Optional[datetime]) -> str:
    return datetime.isoformat(value) if value is not None else None


def optional_iso_datetime_decoder(value: Optional[str]) -> datetime:
    return isoparse(value) if value is not None else None


def datetime_field():
    return field(
        default=None,
        metadata=config(
            encoder=optional_iso_datetime_encoder,
            decoder=optional_iso_datetime_decoder,
            mm_field=fields.DateTime(format='iso')
        )
    )


class ExtendedDtoSerializer(DataClassJsonMixin):
    def to_dict_without_none_values(self):
        d = self.to_dict()

        # Remove values from dict that are None
        return {k: v for k, v in d.items() if v is not None}



import datetime
from dataclasses import dataclass
from typing import Collection, Dict, Optional
from . import ArtifactFile, Model, UserRef


@dataclass
class Artifact(object):
    created_at: datetime = None
    created_by: Optional[UserRef] = None
    files: Optional[Collection[ArtifactFile]] = None
    metadata: Optional[Dict[str, str]] = None
    model: Optional[Model] = None
    name: Optional[str] = None
    run_key: Optional[int] = None
    run_name: Optional[str] = None
    type: Optional[str] = None
    updated_at: datetime = None
    version: Optional[int] = None



from mlaide.error import *
from httpx import Response


def assert_response_status(response: Response, is_404_valid: bool = False):
    if response.status_code == 404 and is_404_valid:
        return

    elif response.status_code == 400:
        raise InputError(_http_error(response))

    elif response.status_code == 401:
        raise InvalidAuthorizationError(_http_error(response))

    elif response.status_code == 403:
        raise NotAuthorizedError(_http_error(response))

    elif response.status_code == 404:
        raise NotFoundError(_http_error(response))

    elif response.status_code >= 500:
        raise ServerError(_http_error(response))


def _http_error(response: Response):
    return HttpError.from_response(response)



from typing import Any, cast, Dict, Optional

import httpx

from ._api_commons import assert_response_status
from ..client import Client
from ..dto import ExperimentDto


def create_experiment(*, client: Client, project_key: str, experiment: ExperimentDto) -> ExperimentDto:

    url = "{}/projects/{projectKey}/experiments".format(client.base_url, projectKey=project_key)

    headers: Dict[str, Any] = client.get_headers()

    json_body = experiment.to_dict_without_none_values()

    response = httpx.post(url=url, headers=headers, json=json_body)

    assert_response_status(response)

    return ExperimentDto.from_dict(cast(Dict[str, Any], response.json()))


def get_experiment(*, client: Client,
                   project_key: str,
                   experiment_key: str) -> Optional[ExperimentDto]:
    url = "{}/projects/{projectKey}/experiments/{experimentKey}" \
        .format(client.base_url, projectKey=project_key, experimentKey=experiment_key)

    headers: Dict[str, Any] = client.get_headers()

    response = httpx.request(
        method="GET",
        url=url,
        headers=headers,
    )

    assert_response_status(response, is_404_valid=True)

    if response.status_code == 404:
        return None

    return ExperimentDto.from_dict(cast(Dict[str, Any], response.json()))



from dataclasses import dataclass
from dataclasses_json import dataclass_json, LetterCase, Undefined

from .run_dto import RunDto


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ExperimentsDto:
    items: RunDto



""" Contains all the data models used in inputs/outputs """

from .artifact_dto import ArtifactDto
from .artifact_ref_dto import ArtifactRefDto
from .experiment_dto import ExperimentDto, ExperimentStatusDto
from .experiment_ref_dto import ExperimentRefDto
from .run_dto import RunDto
from .runs_dto import ExperimentsDto
from .status_dto import StatusDto



import io
import cloudpickle


def serialize(model) -> io.BytesIO:
    for class_obj in model.__class__.__mro__:
        module_name = class_obj.__module__
        if not module_name:
            continue
        # elif module_name.startswith("torch"):
        #     model_type = "torch"
        #     bytestream, method = ensure_bytestream(model)
        #     break
        elif module_name.startswith("sklearn"):
            buffer = io.BytesIO()
            cloudpickle.dump(model, buffer)
            buffer.seek(0)
            return buffer
        # elif module_name.startswith("xgboost"):
        #     model_type = "xgboost"
        #     bytestream, method = ensure_bytestream(model)
        #     break
        # elif module_name.startswith("tensorflow.python.keras"):
        #     model_type = "tensorflow"
        #     tempf = tempfile.NamedTemporaryFile()
        #     if get_tensorflow_major_version() == 2:  # save_format param may not exist in TF 1.X
        #         model.save(tempf.name, save_format='h5')  # TF 2.X uses SavedModel by default
        #     else:
        #         model.save(tempf.name)
        #     tempf.seek(0)
        #     bytestream = tempf
        #     method = "keras"
        #     break


def deserialize(model: io.BytesIO) -> any:
    return cloudpickle.load(model)



from httpx import Response
from .dto import Error


class ApiResponseError(Exception):
    """ An exception raised when an unknown response occurs """

    def __init__(self, *, response: Response, error: Error = None):
        super().__init__()
        self.response: Response = response
        self.error: Error = error




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mlaide',
 'mlaide._api_client',
 'mlaide._api_client.api',
 'mlaide._api_client.dto',
 'mlaide.model']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=1.6.0,<2.0.0',
 'dataclasses-json>=0.5.2,<0.6.0',
 'httpx>=0.16.1,<0.17.0',
 'marshmallow>=3.10.0,<4.0.0',
 'python-dateutil>=2.8.1,<3.0.0']

setup_kwargs = {
    'name': 'mlaide',
    'version': '0.1.0',
    'description': '',
    'long_description': '# Python Client Library for ML Aide\n[![CI pipeline](https://github.com/MLAide/python-client/actions/workflows/ci-pipeline.yml/badge.svg)](https://github.com/MLAide/python-client/actions/workflows/ci-pipeline.yml) [![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=MLAide_python-client&metric=alert_status)](https://sonarcloud.io/dashboard?id=MLAide_python-client)\n\n## Run example scripts\nIn `examples/` you can find some python scripts as an example.\nThe shell scripts can be used for easier usage.\n\nRun the following commands once in a terminal session\n```shell script\ncs/\nsource ./set_api_key.sh\n```\n\nAfter that you can use the following command to run the scripts\nover and over again:\n```shell script\n./run_example.sh\n```\n\n## Contribution\n### Prerequisites\n1. Install [Python](https://www.python.org/)\n2. Install [Python Poetry](https://python-poetry.org/docs/#installation)\n3. Optional - Install IDE: [PyCharm](https://www.jetbrains.com/pycharm/) \nor [Visual Studio Code](https://code.visualstudio.com/)\n\n### Setup Environment\n1. Install environment and download dependencies\n    ```shell\n   poetry install\n   ```\n   \n2. Activate environment\n    ```shell\n    poetry shell\n    ```\n\n### Run Tests\n```\npytest\n```\n\n### Run Tests with Coverage\n```\ncoverage run --branch --source mlaide -m pytest\ncoverage html\n```\n\n### Build\n```\npoetry build\n```\n\n### Publish (on PyPI)\n```\npoetry publish\n```\n\n## Links\n\n- **Homepage:** https://mlaide.com\n- **Quickstart:** https://docs.mlaide.com/start/quickstart/\n- **Tutorial:** https://docs.mlaide.com/tutorial/introduction/\n- **Documentation:** https://docs.mlaide.com/\n',
    'author': 'Raman Singh',
    'author_email': 'mail.raman.s@gmail.com',
    'maintainer': 'Raman Singh',
    'maintainer_email': 'mail.raman.s@gmail.com',
    'url': 'https://mlaide-ai.web.app/home',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



from dataclasses import dataclass
from dataclasses_json import DataClassJsonMixin, config, LetterCase, Undefined
import datetime
from typing import Any, Dict, Optional, List

from .artifact_ref_dto import ArtifactRefDto
from .experiment_ref_dto import ExperimentRefDto
from .helper import datetime_field, ExtendedDtoSerializer
from .status_dto import StatusDto


@dataclass
class RunDto(ExtendedDtoSerializer, DataClassJsonMixin):
    dataclass_json_config = config(
        letter_case=LetterCase.CAMEL,
        undefined=Undefined.EXCLUDE
    )['dataclasses_json']

    created_at: Optional[datetime.datetime] = datetime_field()
    created_by: Optional[Dict[Any, Any]] = None
    end_time: Optional[datetime.datetime] = datetime_field()
    experiment_refs: Optional[List[ExperimentRefDto]] = None
    key: Optional[int] = None
    metrics: Optional[Dict[str, Any]] = None
    name: Optional[str] = None
    parameters: Optional[Dict[str, Any]] = None
    start_time: Optional[datetime.datetime] = datetime_field()
    status: Optional[StatusDto] = None
    used_artifacts: Optional[List[ArtifactRefDto]] = None



from dataclasses import dataclass
from dataclasses_json import DataClassJsonMixin, config, dataclass_json, LetterCase, Undefined
from datetime import datetime
from typing import Any, Dict, List, Optional

from .helper import datetime_field, ExtendedDtoSerializer


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.EXCLUDE)
@dataclass
class ArtifactFileDto:
    file_id: str
    file_name: str


@dataclass
class ModelRevisionDto(object):
    new_stage: str
    old_stage: str
    created_at: Optional[datetime] = datetime_field()
    created_by: Optional[Dict[Any, Any]] = None
    note: Optional[str] = None


@dataclass
class ModelDto(object):
    created_at: Optional[datetime] = datetime_field()
    created_by: Optional[Dict[Any, Any]] = None
    model_revisions: Optional[List[ModelRevisionDto]] = None
    metadata: Optional[Dict[str, str]] = None
    stage: Optional[str] = None
    updated_at: Optional[datetime] = datetime_field()


@dataclass
class ArtifactDto(ExtendedDtoSerializer, DataClassJsonMixin):
    dataclass_json_config = config(
        letter_case=LetterCase.CAMEL,
        undefined=Undefined.EXCLUDE
    )['dataclasses_json']

    created_at: Optional[datetime] = datetime_field()
    created_by: Optional[Dict[Any, Any]] = None
    files: Optional[List[ArtifactFileDto]] = None
    metadata: Optional[Dict[str, str]] = None
    model: Optional[ModelDto] = None
    name: Optional[str] = None
    run_key: Optional[int] = None
    run_name: Optional[str] = None
    type: Optional[str] = None
    updated_at: Optional[datetime] = datetime_field()
    version: Optional[int] = None
